module ISet = Set.Make (Int)

type t = {
  init : ISet.t;
  final : ISet.t;
  delta : (int, (char, ISet.t) Hashtbl.t) Hashtbl.t;
  cache : (ISet.t * char, ISet.t) Hashtbl.t;
}

let make init final trs =
  let delta = Hashtbl.create 16 in
  List.iter
    (fun (q, a, p) ->
      let t =
        try Hashtbl.find delta q
        with Not_found ->
          let t = Hashtbl.create 16 in
          Hashtbl.add delta q t;
          t
      in
      let s = try Hashtbl.find t a with Not_found -> ISet.empty in
      Hashtbl.replace t a (ISet.add p s))
    trs;
  { init; final; delta; cache = Hashtbl.create 16 }

let next_state auto q a =
  try Hashtbl.find (Hashtbl.find auto.delta q) a with Not_found -> ISet.empty

let naive_run auto txt =
  let rec loop q i =
    if i >= String.length txt then ISet.mem q auto.final
    else ISet.exists (fun p -> loop p (i + 1)) (next_state auto q txt.[i])
  in
  ISet.exists (fun q -> loop q 0) auto.init

let next auto s a =
  ISet.fold (fun q acc -> ISet.union acc (next_state auto q a)) s ISet.empty

let next_cache auto s a =
  try Hashtbl.find auto.cache (s, a)
  with Not_found ->
    let res = next auto s a in
    Hashtbl.add auto.cache (s, a) res;
    res

let run auto txt =
  let rec loop s i =
    if i >= String.length txt || ISet.is_empty s then
      not (ISet.disjoint auto.final s)
    else loop (next_cache auto s txt.[i]) (i + 1)
  in
  loop auto.init 0

let auto0 =
  make (ISet.of_list [ 0 ]) (ISet.of_list [ 2 ])
    ([ (0, 'a', 0); (0, 'a', 1); (1, 'a', 0); (1, 'b', 2) ]
    @ [ (0, 'a', 3); (3, 'a', 3) ])

let test () =
  for i = 0 to 10000 do
    let txt = String.make i 'a' in
    let t0 = Unix.gettimeofday () in
    let _ = run auto0 txt in
    let t1 = Unix.gettimeofday () in
    Format.printf "%d : %f\n%!" i (t1 -. t0)
  done

let () = test ()
